import json
from dataclasses import dataclass
from functools import reduce
from pathlib import Path
from typing import Any, Dict, Iterable, Optional, Set

VOCAB_FILE = "vocab.json"


@dataclass
class Vocab:
    """A class which represents a dictionary of encountered tokens in a dataset."""

    vocab: Dict[str, int]

    @property
    def symbols(self) -> Set[str]:
        return set(self.vocab.keys())

    def merge(self, other: "Vocab") -> "Vocab":
        """Creates a new Vocab which includes all symbols in the merged two."""
        vocab = self.symbols | other.symbols
        return Vocab.from_set(vocab)

    def save(self, path: Path) -> None:
        """Saves the vocab to the supplied path.

        If the path is a folder, saves as vocab.json, within it.
        """
        if path.is_dir():
            path /= VOCAB_FILE

        with open(path, "w") as out:
            json.dump(self.vocab, out)

    def add(self, char: str) -> None:
        """Adds a new character into the vocab."""
        if char in self.vocab:
            return

        self.vocab[char] = len(self.vocab)

    def replace(self, original: str, replacement: str) -> None:
        """Replaces the supplied character mapping in the vocab."""
        if original not in self.vocab:
            return

        self.vocab[replacement] = self.vocab[original]
        self.vocab.pop(original)

    @classmethod
    def from_set(cls, symbols: Set[str]) -> "Vocab":
        """Builds a vocab from a set of symbols."""
        vocab = {symbol: index for index, symbol in enumerate(sorted(symbols))}
        return cls(vocab=vocab)

    @classmethod
    def from_strings(cls, texts: Iterable[str]) -> "Vocab":
        """Builds an vocab from a iterable text collection."""

        def reducer(result: Set[str], text: str) -> Set[str]:
            return result | set(text)

        symbols = reduce(reducer, texts, set())
        return cls.from_set(symbols)



import os
from itertools import chain
from pathlib import Path
from typing import Any, Dict, List, Optional

import numpy as np
from datasets import Audio, DatasetDict, load_dataset
from loguru import logger
from transformers import Wav2Vec2Processor

PROCESSOR_COUNT = 4
AUDIO_COLUMN = "audio"
SAMPLING_RATE = 16_000


def create_dataset(
    dataset_path: Path, cache_dir: Optional[Path] = None, test_size: float = 0.2
) -> DatasetDict:
    """Creates a dataset with test/train splits from the data within a given
    directory.

    Parameters:
        dataset_path: The path to the unprocessed dataset files.
        cache_dir: The path to save the processed dataset.
        test_size: The percentage of the dataset to allocate as the test set.

    Returns:
        A dataset dictionary with test and train splits.
    """
    transcript_files = [
        str(dataset_path / file)
        for file in os.listdir(dataset_path)
        if (dataset_path / file).suffix == ".json"
    ]
    logger.debug(f"Transcript file paths sample: {transcript_files[:4]}")

    # Annoying hack
    if cache_dir is not None:
        cache_dir = str(cache_dir)  # type: ignore

    dataset = load_dataset("json", data_files=transcript_files, cache_dir=cache_dir)  # type: ignore

    # Convert the audio file name column into the matching audio data
    dataset = dataset.rename_column("audio_file", AUDIO_COLUMN)
    logger.debug(f"Dataset audio file paths sample: {dataset['train'][AUDIO_COLUMN][:4]}")  # type: ignore

    def resolve_audio_path(row: Dict[str, Any]) -> Dict[str, Any]:
        # Forcefully resolve to same dir as dataset.
        path = dataset_path / Path(row[AUDIO_COLUMN]).name
        row[AUDIO_COLUMN] = str(path)
        return row

    dataset = dataset.map(resolve_audio_path)
    logger.debug(f"Dataset audio file paths post-resolution: {dataset['train'][AUDIO_COLUMN][:4]}")  # type: ignore
    dataset = dataset.cast_column(AUDIO_COLUMN, Audio(sampling_rate=SAMPLING_RATE))

    return dataset["train"].train_test_split(test_size=test_size)  # type: ignore


def prepare_dataset(dataset: DatasetDict, processor: Wav2Vec2Processor) -> DatasetDict:
    """Runs some preprocessing over the given dataset.

    TODO: I'm going to be honest, I have no idea what this does, and need some
    smart ML knight in shining armour to write a propert description.

    Parameters:
        dataset: The dataset to apply the preprocessing
        processor: The processor to apply over the dataset
    """

    logger.debug(f"Dataset pre prep: {dataset}")
    logger.debug(f"Dataset[train] pre prep: {dataset['train']['transcript'][0]}")
    logger.debug(
        f'Input array shape:, {np.asarray(dataset["train"][0]["audio"]["array"]).shape}'
    )
    logger.debug(f'Sampling rate:, {dataset["train"][0]["audio"]["sampling_rate"]}')
    logger.debug(f"Tokenizer vocab: {processor.tokenizer.vocab}")  # type: ignore

    def _prepare_dataset(batch: Dict) -> Dict[str, List]:
        # Also from https://huggingface.co/blog/fine-tune-xlsr-wav2vec2
        audio = batch["audio"]

        batch["input_values"] = processor(
            audio["array"], sampling_rate=audio["sampling_rate"]
        ).input_values[0]
        batch["input_length"] = len(batch["input_values"])

        batch["labels"] = processor(text=batch["transcript"]).input_ids

        return batch

    columns = dataset.column_names.values()
    # flatten and make unique between datasets
    columns_to_remove = list(set(chain.from_iterable(columns)))

    dataset = dataset.map(
        _prepare_dataset,
        remove_columns=columns_to_remove,
        num_proc=PROCESSOR_COUNT,
    )

    logger.debug(f"Dataset post prep: {dataset}")
    logger.debug(f"Training labels: {dataset['train']['labels'][0]}")
    # logger.debug(f"Training inputs: {dataset['train']['input_values'][0]}")
    return dataset



from contextlib import nullcontext
from pathlib import Path
from typing import Optional

from datasets import DatasetDict
from loguru import logger
from transformers import (
    AutoConfig,
    AutoFeatureExtractor,
    AutoModelForCTC,
    AutoProcessor,
    AutoTokenizer,
    Trainer,
    Wav2Vec2CTCTokenizer,
    Wav2Vec2FeatureExtractor,
    Wav2Vec2Processor,
)

from elpis.datasets import create_dataset, prepare_dataset
from elpis.models.vocab import VOCAB_FILE, Vocab
from elpis.trainer.data_collator import DataCollatorCTCWithPadding
from elpis.trainer.job import TrainingJob
from elpis.trainer.metrics import create_metrics
from elpis.trainer.utils import log_to_file


def create_processor(
    job: TrainingJob,
    output_dir: Path,
    dataset: DatasetDict,
    cache_dir: Optional[Path],
    unk_token="[UNK]",
    pad_token="[PAD]",
    word_delimiter_token="|",
) -> Wav2Vec2Processor:
    config = AutoConfig.from_pretrained(job.base_model)
    tokenizer_type = config.model_type if config.tokenizer_class is None else None
    config = config if config.tokenizer_class is not None else None

    # Build up a vocab from the dataset.
    train_vocab = Vocab.from_strings(dataset["train"]["transcript"])
    test_vocab = Vocab.from_strings(dataset["test"]["transcript"])
    vocab = train_vocab.merge(test_vocab)

    vocab.add(unk_token)
    vocab.add(pad_token)
    vocab.replace(" ", word_delimiter_token)  # feels a little restrictive?
    logger.info(f"Vocab: {vocab.vocab}")
    vocab.save(output_dir)

    tokenizer = AutoTokenizer.from_pretrained(
        output_dir,
        config=config,
        tokenizer_type=tokenizer_type,
        unk_token=unk_token,
        pad_token=pad_token,
        word_delimiter_token=word_delimiter_token,
        cache_dir=cache_dir,
    )

    feature_extractor = AutoFeatureExtractor.from_pretrained(
        job.base_model, cache_dir=cache_dir
    )

    return Wav2Vec2Processor(feature_extractor=feature_extractor, tokenizer=tokenizer)


def train(
    job: TrainingJob,
    output_dir: Path,
    dataset_dir: Path,
    cache_dir: Optional[Path] = None,
    log_file: Optional[Path] = None,
) -> Path:
    """Fine-tunes a model for use in transcription.

    Parameters:
        job: Info about the training job, e.g. training options.
        output_dir: Where to save the trained model.
        dataset_dir: A directory containing the preprocessed dataset to train with.
        cache_dir: A directory to use for caching HFT downloads and datasets.
        log_file: An optional file to write training logs to.

    Returns:
        A path to the folder containing the trained model.
    """

    context = log_to_file(log_file) if log_file is not None else nullcontext()
    with context:
        logger.info("Preparing Datasets...")
        dataset = create_dataset(dataset_dir, cache_dir)
        processor = create_processor(job, output_dir, dataset, cache_dir)
        dataset = prepare_dataset(dataset, processor)
        logger.info("Finished Preparing Datasets")

        logger.info("Downloading pretrained model...")
        model = AutoModelForCTC.from_pretrained(
            job.base_model,
            cache_dir=cache_dir,
            ctc_loss_reduction="mean",
            pad_token_id=processor.tokenizer.pad_token_id,  # type: ignore
            # Wav2vec2 specific hyperparams copied from docs.
            attention_dropout=0.1,
            hidden_dropout=0.1,
            feat_proj_dropout=0.0,
            mask_time_prob=0.05,
            layerdrop=0.1,
            vocab_size=len(processor.tokenizer),  # type: ignore
            # For Ash -> errors if below param not set.
            ignore_mismatched_sizes=True,
        )
        logger.info("Downloaded model.")

        if job.options.freeze_feature_extractor:
            model.freeze_feature_extractor()

        data_collator = DataCollatorCTCWithPadding(processor=processor, padding=True)
        output_dir.mkdir(exist_ok=True, parents=True)

        trainer = Trainer(
            model=model,
            args=job.to_training_args(output_dir),
            train_dataset=dataset["train"],  # type: ignore
            eval_dataset=dataset["test"],  # type: ignore
            tokenizer=processor.feature_extractor,  # type: ignore
            data_collator=data_collator,
            compute_metrics=create_metrics(job.metrics, processor),
        )

        logger.info(f"Begin training model...")
        trainer.train()
        logger.info(f"Finished training!")

        logger.info(f"Saving model @ {output_dir}")
        trainer.save_model()
        trainer.save_state()
        processor.save_pretrained(output_dir)
        logger.info(f"Model written to disk.")

        metrics = trainer.evaluate()
        logger.info("==== Metrics ====")
        trainer.log_metrics("eval", metrics)
        trainer.save_metrics("eval", metrics)
        logger.info(metrics)

        return output_dir



from typing import Callable, Dict, Optional, Sequence

import evaluate
import numpy as np
from loguru import logger
from transformers import EvalPrediction, Wav2Vec2Processor


def create_metrics(
    metric_names: Sequence[str], processor: Wav2Vec2Processor
) -> Optional[Callable[[EvalPrediction], Dict]]:
    # Handle metrics
    if len(metric_names) == 0:
        return

    # Note: was using evaluate.combine but was having many unexpected errors.
    metrics = {name: evaluate.load(name) for name in metric_names}

    def compute_metrics(pred: EvalPrediction) -> Dict:
        # taken from https://huggingface.co/blog/fine-tune-xlsr-wav2vec2
        pred_logits = pred.predictions

        pred.label_ids[pred.label_ids == -100] = processor.tokenizer.pad_token_id  # type: ignore

        # Taken from: https://discuss.huggingface.co/t/code-review-compute-metrics-for-wer-with-wav2vec2processorwithlm/16841/3
        if type(processor).__name__ == "Wav2Vec2ProcessorWithLM":
            pred_str = processor.batch_decode(pred_logits).text
        else:
            pred_ids = np.argmax(pred_logits, axis=-1)
            pred_str = processor.batch_decode(pred_ids)

        # We do not want to group tokens when computing the metrics
        label_str = processor.batch_decode(pred.label_ids, group_tokens=False)

        logger.debug(f"METRICS->pred: {pred_str} label:{label_str}")

        result = {
            name: metric.compute(predictions=pred_str, references=label_str)
            for name, metric in metrics.items()
        }
        logger.debug(f"Metrics Result: {result}")
        return result

    return compute_metrics



from elpis.models.annotation import Annotation
from elpis.models.elan_options import ElanOptions, ElanTierSelector
from elpis.models.vocab import VOCAB_FILE, Vocab

__all__ = ["Annotation", "ElanOptions", "ElanTierSelector", "Vocab", "VOCAB_FILE"]



import logging
import sys
from contextlib import contextmanager
from logging.handlers import WatchedFileHandler
from pathlib import Path

from loguru import logger
from transformers.utils import logging as transformer_logging


@contextmanager
def log_to_file(log_file: Path):
    """A context manager which logs its captured runtime to the provided file."""
    sink_id = logger.add(log_file)

    handler = WatchedFileHandler(str(log_file))
    formatter = logging.Formatter(
        fmt="%(asctime)s.%(msecs)03d | %(levelname)s | (HF) %(name)s |   %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    handler.setFormatter(formatter)
    handler.setLevel(logging.INFO)

    logging.root.addHandler(handler)

    # Propagate huggingface logs to logging root.
    transformer_logging.set_verbosity_info()
    transformer_logging.enable_propagation()
    transformer_logging.disable_default_handler()

    # Log stderr to log file to capture tqdm logs
    with open(log_file, "a") as stderr_hole:
        original_stderr = sys.stderr
        try:
            sys.stderr = stderr_hole
            yield log_file
        finally:
            sys.stderr = original_stderr
            # Flush and teardown logging handlers
            logger.remove(sink_id)
            handler.flush()
            logging.root.removeHandler(handler)
